"""Classes to handle encoding of Documents on DB for the Vector Store.."""

from __future__ import annotations

import logging
import warnings
from abc import ABC, abstractmethod
from typing import Any, Iterable

from langchain_core.documents import Document
from typing_extensions import override

NO_NULL_VECTOR_MSG = "Default (non-vectorize) codec cannot encode null vector."
VECTOR_REQUIRED_PREAMBLE_MSG = (
    "Default vectorize codec got a non-null vector to encode."
)
FLATTEN_CONFLICT_MSG = "Cannot flatten metadata: field name overlap for '{field}'."
VECTORIZE_NOT_AVAILABLE_MSG = "Vectorize not available for this codec."

VECTOR_FIELD_NAME = "$vector"
VECTORIZE_FIELD_NAME = "$vectorize"
LEXICAL_FIELD_NAME = "$lexical"
SIMILARITY_FIELD_NAME = "$similarity"
RERANK_SORT_TOP_FIELD_NAME = "$hybrid"
DEFAULT_METADATA_FIELD_NAME = "metadata"

STANDARD_INDEXING_OPTIONS_DEFAULT = {"allow": [DEFAULT_METADATA_FIELD_NAME]}

logger = logging.getLogger(__name__)


def _default_decode_vector(astra_doc: dict[str, Any]) -> list[float] | None:
    """Extract the embedding vector from an Astra DB document."""
    return astra_doc.get(VECTOR_FIELD_NAME)


def _default_metadata_key_to_field_identifier(md_key: str) -> str:
    """Rewrite a metadata key name to its full path in the 'default' encoding.

    The input `md_key` is an "abstract" metadata key, while the return value
    identifies its actual full-path location on an Astra DB document encoded in the
    'default' way (i.e. with a nested `metadata` dictionary).
    """
    return f"{DEFAULT_METADATA_FIELD_NAME}.{md_key}"


def _flat_metadata_key_to_field_identifier(md_key: str) -> str:
    """Rewrite a metadata key name to its full path in the 'flat' encoding.

    The input `md_key` is an "abstract" metadata key, while the return value
    identifies its actual full-path location on an Astra DB document encoded in the
    'flat' way (i.e. metadata fields appearing at top-level in the Astra DB document).
    """
    return md_key


def _default_encode_filter(filter_dict: dict[str, Any]) -> dict[str, Any]:
    """Encode an "abstract" filter/sort condition for the 'default' encoding.

    The input can express a query clause, or sort criterion, on metadata and uses
    just the metadata field names, possibly connected/nested through AND and ORs.
    The output makes key names into their full path-identifiers (e.g. "metadata.xyz")
    according to the 'default' encoding scheme for Astra DB documents.
    """
    metadata_filter = {}
    for k, v in filter_dict.items():
        # Key in this dict starting with $ are supposedly operators and as such
        # should not be nested within the `metadata.` prefix. For instance,
        # >>> _default_encode_filter({'a':1, '$or': [{'b':2}, {'c': 3}]})
        #     {'metadata.a': 1, '$or': [{'metadata.b': 2}, {'metadata.c': 3}]}
        if k and k[0] == "$":
            if isinstance(v, list) and k != VECTOR_FIELD_NAME:
                metadata_filter[k] = [_default_encode_filter(f) for f in v]
            elif isinstance(v, dict):
                # assume each list item can be fed back to this function
                metadata_filter[k] = _default_encode_filter(v)  # type: ignore[assignment]
            else:
                # a scalar. As this is a 'value', never touch it
                metadata_filter[k] = v
        else:
            metadata_filter[_default_metadata_key_to_field_identifier(k)] = v

    return metadata_filter


def _astra_generic_encode_id(filter_id: str) -> dict[str, Any]:
    """Encoding of a single Document ID as a query clause for an Astra DB document."""
    return {"_id": filter_id}


def _astra_generic_encode_ids(filter_ids: list[str]) -> dict[str, Any]:
    """Encoding of Document IDs as a query clause for an Astra DB document.

    This function picks the right, and most concise, expression based on the
    multiplicity of the provided IDs.
    """
    if len(filter_ids) == 1:
        return _astra_generic_encode_id(filter_ids[0])
    return {"_id": {"$in": filter_ids}}


def _astra_generic_encode_vector_sort(vector: list[float]) -> dict[str, Any]:
    """Encoding of a vector-based sort as a query clause for an Astra DB document."""
    return {VECTOR_FIELD_NAME: vector}


def _astra_generic_encode_vectorize_sort(query_text: str) -> dict[str, Any]:
    """Encoding of a vectorize-based sort as a query clause for an Astra DB document."""
    return {VECTORIZE_FIELD_NAME: query_text}


def _astra_vector_encode_hybrid_sort(
    vector: list[float],
    lexical: str,
) -> dict[str, Any]:
    """Encoding of a sort clause for hybrid search in the non-vectorize case."""
    return {
        RERANK_SORT_TOP_FIELD_NAME: {
            VECTOR_FIELD_NAME: vector,
            LEXICAL_FIELD_NAME: lexical,
        },
    }


def _astra_vectorize_encode_hybrid_sort(
    vectorize: str,
    lexical: str,
) -> dict[str, Any]:
    """Encoding of a sort clause for hybrid search in the vectorize case."""
    if vectorize == lexical:
        return {RERANK_SORT_TOP_FIELD_NAME: vectorize}
    return {
        RERANK_SORT_TOP_FIELD_NAME: {
            VECTORIZE_FIELD_NAME: vectorize,
            LEXICAL_FIELD_NAME: lexical,
        },
    }


class _AstraDBVectorStoreDocumentCodec(ABC):
    """A Document codec for the Astra DB vector store.

    Document codecs hold the logic consistent interaction
    with documents as stored on the Astra DB collection.

    In this context, 'Document' (capital D) refers to the LangChain class,
    while 'Astra DB document' refers to the JSON-like object stored on DB.

    Implementations of this class must:
    - define how to encode/decode documents consistently to and from
      Astra DB collections. The two operations must, so to speak, combine
      to the identity on both sides (except for the quirks of their signatures).
    - provide the adequate projection dictionaries for running find
      operations on Astra DB, with and without the field containing the vector.
    - encode Document IDs to the right field on Astra DB ("_id" for Collections).
    - define the name of the field storing the textual content of the Document.
    - define whether embeddings are computed server-side (with $vectorize) or not.
    """

    server_side_embeddings: bool
    has_lexical: bool
    content_field: str
    base_projection: dict[str, bool]
    full_projection: dict[str, bool]
    ignore_invalid_documents: bool

    @abstractmethod
    def encode(
        self,
        content: str,
        document_id: str,
        vector: list[float] | None,
        metadata: dict | None,
    ) -> dict[str, Any]:
        """Create a document for storage on Astra DB.

        Args:
            content: textual content for the (LangChain) `Document`.
            document_id: unique ID for the (LangChain) `Document`.
            vector: a vector associated to the (LangChain) `Document`. This
                parameter must be None for and only for server-side embeddings.
            metadata: a metadata dictionary for the (LangChain) `Document`.

        Returns:
            a dictionary ready for storage onto Astra DB.
        """

    @abstractmethod
    def decode(self, astra_document: dict[str, Any]) -> Document | None:
        """Create a LangChain Document instance from a document retrieved from Astra DB.

        Args:
            astra_document: a dictionary as retrieved from Astra DB.

        Returns:
            a (langchain) Document corresponding to the input.
        """

    @abstractmethod
    def encode_filter(self, filter_dict: dict[str, Any]) -> dict[str, Any]:
        """Encode a LangChain filter for use in Astra DB queries.

        Make a LangChain filter into a filter clause suitable for operations
        against the Astra DB collection, consistently with the encoding scheme.

        Args:
            filter_dict: a filter in the standardized metadata-filtering form
                used throughout LangChain.

        Returns:
            an equivalent filter clause for use in Astra DB's find queries.
        """

    @abstractmethod
    def metadata_key_to_field_identifier(self, md_key: str) -> str:
        """Express an 'abstract' metadata key as a full Data API field identifier."""

    @abstractmethod
    def encode_vectorize_sort(self, query_text: str) -> dict[str, Any]:
        """Encode a query text as a sort to use for Astra DB 'vectorize' queries.

        Note that calling this method on codecs that don't support 'vectorize'
        will result in an exception being thrown.

        Args:
            query_text: the search query text to order results by.

        Returns:
            an order clause for use in Astra DB's find queries.
        """

    @abstractmethod
    def encode_hybrid_sort(
        self,
        *,
        vector: list[float] | None,
        vectorize: str | None,
        lexical: str,
    ) -> dict[str, Any]:
        """Encode a 'sort' parameter for an Astra DB 'hybrid' search.

        The input parameters must be appropriate for the particular codec:
        supplying the wrong inputs (such as a vector on a vectorize-codec)
        will result in an exception being thrown.

        Args:
            vector: a query vector (if applicable) or None.
            vectorize: a query text for vectorize search (if applicable) or None.
                Exactly one of ``vector`` and ``vectorize`` must not be None.
            lexical: the search query text for the lexical part of the hybrid search.

        Returns:
            a sort clause for use in Astra DB's findAndRerank queries.
        """

    @property
    @abstractmethod
    def default_collection_indexing_policy(self) -> dict[str, list[str]]:
        """Provide the default indexing policy if the collection must be created."""

    @property
    def rerank_on(self) -> str | None:
        """The value for 'rerank_on' in a find_and_rerank command, or None.

        This property is not None if and only if the codec is a non-vectorize one.
        """
        if self.server_side_embeddings:
            return None
        return self.content_field

    def decode_vector(self, astra_document: dict[str, Any]) -> list[float] | None:
        """Create a vector from a document retrieved from Astra DB.

        Args:
            astra_document: a dictionary as retrieved from Astra DB.

        Returns:
            a vector corresponding to the input.
        """
        return _default_decode_vector(astra_document)

    def get_id(self, astra_document: dict[str, Any]) -> str:
        """Return the ID of an encoded document (= a raw JSON read from DB)."""
        return astra_document["_id"]

    def get_similarity(self, astra_document: dict[str, Any]) -> float | None:
        """Return the similarity of an encoded document (= a raw JSON read from DB).

        This method gives no guarantees as to whether said score applies/is found.
        """
        return astra_document.get(SIMILARITY_FIELD_NAME)

    def encode_query(
        self,
        *,
        ids: Iterable[str] | None = None,
        filter_dict: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        """Prepare an encoded query according to the Astra DB document encoding.

        The method optionally accepts both IDs and metadata filters. The two,
        if passed together, are automatically combined with an AND operation.

        In other words, if passing both IDs and a metadata filtering clause,
        the resulting query would return Astra DB documents matching the metadata
        clause AND having an ID among those provided to this method. If, instead,
        an OR is required, one should run two separate queries and subsequently merge
        the result (taking care of avoiding duplcates).

        Args:
            ids: an iterable over Document IDs. If provided, the resulting Astra DB
                query dictionary expresses the requirement that returning documents
                have an ID among those provided here. Passing an empty iterable,
                or None, results in a query with no conditions on the IDs at all.
            filter_dict: a metadata filtering part. If provided, if must refer to
                metadata keys by their bare name (such as `{"key": 123}`).
                This filter can combine nested conditions with "$or"/"$and" connectors,
                for example:
                - `{"tag": "a"}`
                - `{"$or": [{"tag": "a"}, "label": "b"]}`
                - `{"$and": [{"tag": {"$in": ["a", "z"]}}, "label": "b"]}`

        Returns:
            a query dictionary ready to be used in an Astra DB find operation on
            a collection.
        """
        clauses: list[dict[str, Any]] = []
        _ids_list = list(ids or [])
        if _ids_list:
            clauses.append(_astra_generic_encode_ids(_ids_list))
        if filter_dict:
            clauses.append(self.encode_filter(filter_dict))

        if clauses:
            if len(clauses) > 1:
                return {"$and": clauses}
            return clauses[0]
        return {}

    def encode_vector_sort(self, vector: list[float]) -> dict[str, Any]:
        """Encode a vector as a sort to use for Astra DB queries.

        Args:
            vector: the search vector to order results by.

        Returns:
            an order clause for use in Astra DB's find queries.
        """
        return _astra_generic_encode_vector_sort(vector)


class _DefaultVSDocumentCodec(_AstraDBVectorStoreDocumentCodec):
    """Codec for the default vector store usage with client-side embeddings.

    This codec expresses how document are stored for collections created
    and entirely managed by the AstraDBVectorStore class.
    """

    server_side_embeddings = False

    def __init__(
        self, content_field: str, *, ignore_invalid_documents: bool, has_lexical: bool
    ) -> None:
        """Initialize a new DefaultVSDocumentCodec.

        Args:
            content_field: name of the (top-level) field for textual content.
            ignore_invalid_documents: if True, noncompliant inputs to `decode`
                are logged and a None is returned (instead of raising an exception).
            has_lexical: whether the codec should use the lexical field (hybrid search)
        """
        self.content_field = content_field
        self.has_lexical = has_lexical
        self.base_projection = {
            "_id": True,
            self.content_field: True,
            DEFAULT_METADATA_FIELD_NAME: True,
        }
        self.full_projection = {
            "_id": True,
            self.content_field: True,
            DEFAULT_METADATA_FIELD_NAME: True,
            VECTOR_FIELD_NAME: True,
        }
        self.ignore_invalid_documents = ignore_invalid_documents

    @override
    def encode(
        self,
        content: str,
        document_id: str,
        vector: list[float] | None,
        metadata: dict | None,
    ) -> dict[str, Any]:
        if vector is None:
            raise ValueError(NO_NULL_VECTOR_MSG)
        return {
            self.content_field: content,
            "_id": document_id,
            VECTOR_FIELD_NAME: vector,
            DEFAULT_METADATA_FIELD_NAME: metadata or {},
            **({LEXICAL_FIELD_NAME: content} if self.has_lexical else {}),
        }

    @override
    def decode(self, astra_document: dict[str, Any]) -> Document | None:
        _invalid_doc = (
            DEFAULT_METADATA_FIELD_NAME not in astra_document
            or self.content_field not in astra_document
        )
        if _invalid_doc and self.ignore_invalid_documents:
            invalid_doc_warning = (
                "Ignoring document with _id = "
                f"{astra_document.get('_id', '(no _id)')}. "
                "Reason: missing required fields."
            )
            logger.warning(invalid_doc_warning)
            return None
        return Document(
            page_content=astra_document[self.content_field],
            metadata=astra_document[DEFAULT_METADATA_FIELD_NAME],
            id=astra_document["_id"],
        )

    @override
    def encode_filter(self, filter_dict: dict[str, Any]) -> dict[str, Any]:
        return _default_encode_filter(filter_dict)

    @override
    def metadata_key_to_field_identifier(self, md_key: str) -> str:
        return _default_metadata_key_to_field_identifier(md_key)

    @override
    def encode_vectorize_sort(self, query_text: str) -> dict[str, Any]:
        raise ValueError(VECTORIZE_NOT_AVAILABLE_MSG)

    @override
    def encode_hybrid_sort(
        self,
        *,
        vector: list[float] | None,
        vectorize: str | None,
        lexical: str,
    ) -> dict[str, Any]:
        if vector is None or vectorize is not None:
            msg = "This codec's hybrid sort requires `vectorize=None` and a vector."
            raise ValueError(msg)
        return _astra_vector_encode_hybrid_sort(vector=vector, lexical=lexical)

    @property
    @override
    def default_collection_indexing_policy(self) -> dict[str, list[str]]:
        return STANDARD_INDEXING_OPTIONS_DEFAULT


class _DefaultVectorizeVSDocumentCodec(_AstraDBVectorStoreDocumentCodec):
    """Codec for the default vector store usage with server-side embeddings.

    This codec expresses how document are stored for collections created
    and entirely managed by the AstraDBVectorStore class, for the case of
    server-side embeddings (aka $vectorize).
    """

    server_side_embeddings = True
    content_field = VECTORIZE_FIELD_NAME

    def __init__(self, *, ignore_invalid_documents: bool, has_lexical: bool) -> None:
        """Initialize a new DefaultVectorizeVSDocumentCodec.

        Args:
            ignore_invalid_documents: if True, noncompliant inputs to `decode`
                are logged and a None is returned (instead of raising an exception).
            has_lexical: whether the codec should use the lexical field (hybrid search)
        """
        self.base_projection = {
            "_id": True,
            VECTORIZE_FIELD_NAME: True,
            DEFAULT_METADATA_FIELD_NAME: True,
        }
        self.full_projection = {
            "_id": True,
            VECTORIZE_FIELD_NAME: True,
            DEFAULT_METADATA_FIELD_NAME: True,
            VECTOR_FIELD_NAME: True,
        }
        self.ignore_invalid_documents = ignore_invalid_documents
        self.has_lexical = has_lexical

    @override
    def encode(
        self,
        content: str,
        document_id: str,
        vector: list[float] | None,
        metadata: dict | None,
    ) -> dict[str, Any]:
        if vector is not None:
            msg = f"{VECTOR_REQUIRED_PREAMBLE_MSG}: {vector}"
            raise ValueError(msg)
        return {
            VECTORIZE_FIELD_NAME: content,
            "_id": document_id,
            DEFAULT_METADATA_FIELD_NAME: metadata or {},
            **({LEXICAL_FIELD_NAME: content} if self.has_lexical else {}),
        }

    @override
    def decode(self, astra_document: dict[str, Any]) -> Document | None:
        _invalid_doc = (
            DEFAULT_METADATA_FIELD_NAME not in astra_document
            or VECTORIZE_FIELD_NAME not in astra_document
        )
        if _invalid_doc and self.ignore_invalid_documents:
            invalid_doc_warning = (
                "Ignoring document with _id = "
                f"{astra_document.get('_id', '(no _id)')}. "
                "Reason: missing required fields."
            )
            warnings.warn(
                invalid_doc_warning,
                stacklevel=2,
            )
            return None
        return Document(
            page_content=astra_document[VECTORIZE_FIELD_NAME],
            metadata=astra_document[DEFAULT_METADATA_FIELD_NAME],
            id=astra_document["_id"],
        )

    @override
    def encode_filter(self, filter_dict: dict[str, Any]) -> dict[str, Any]:
        return _default_encode_filter(filter_dict)

    @override
    def metadata_key_to_field_identifier(self, md_key: str) -> str:
        return _default_metadata_key_to_field_identifier(md_key)

    @override
    def encode_vectorize_sort(self, query_text: str) -> dict[str, Any]:
        return _astra_generic_encode_vectorize_sort(query_text)

    @override
    def encode_hybrid_sort(
        self,
        *,
        vector: list[float] | None,
        vectorize: str | None,
        lexical: str,
    ) -> dict[str, Any]:
        if vectorize is None or vector is not None:
            msg = "This codec's hybrid sort requires `vector=None` and a vectorize."
            raise ValueError(msg)
        return _astra_vectorize_encode_hybrid_sort(vectorize=vectorize, lexical=lexical)

    @property
    @override
    def default_collection_indexing_policy(self) -> dict[str, list[str]]:
        return STANDARD_INDEXING_OPTIONS_DEFAULT


class _FlatVSDocumentCodec(_AstraDBVectorStoreDocumentCodec):
    """Codec for collections populated externally, with client-side embeddings.

    This codec manages document structured as a flat key-value map, with one
    field being the textual content and the other implicitly forming the "metadata".
    """

    server_side_embeddings = False

    def __init__(
        self, content_field: str, *, ignore_invalid_documents: bool, has_lexical: bool
    ) -> None:
        """Initialize a new DefaultVSDocumentCodec.

        Args:
            content_field: name of the (top-level) field for textual content.
            ignore_invalid_documents: if True, noncompliant inputs to `decode`
                are logged and a None is returned (instead of raising an exception).
            has_lexical: whether the codec should use the lexical field (hybrid search)
        """
        self.content_field = content_field
        self.base_projection = {"_id": True, VECTOR_FIELD_NAME: False}
        self.full_projection = {"*": True}
        self.ignore_invalid_documents = ignore_invalid_documents
        self.has_lexical = has_lexical
        self._non_md_fields = {
            "_id",
            self.content_field,
            LEXICAL_FIELD_NAME,
            VECTOR_FIELD_NAME,
            VECTORIZE_FIELD_NAME,
            SIMILARITY_FIELD_NAME,
        }

    @override
    def encode(
        self,
        content: str,
        document_id: str,
        vector: list[float] | None,
        metadata: dict | None,
    ) -> dict[str, Any]:
        if vector is None:
            raise ValueError(NO_NULL_VECTOR_MSG)
        if self.content_field in (metadata or {}):
            msg = FLATTEN_CONFLICT_MSG.format(field=self.content_field)
            raise ValueError(msg)
        return {
            self.content_field: content,
            "_id": document_id,
            VECTOR_FIELD_NAME: vector,
            **(metadata or {}),
            **({LEXICAL_FIELD_NAME: content} if self.has_lexical else {}),
        }

    @override
    def decode(self, astra_document: dict[str, Any]) -> Document | None:
        if self.content_field not in astra_document and self.ignore_invalid_documents:
            invalid_doc_warning = (
                "Ignoring document with _id = "
                f"{astra_document.get('_id', '(no _id)')}. "
                "Reason: missing required fields."
            )
            warnings.warn(
                invalid_doc_warning,
                stacklevel=2,
            )
            return None
        _metadata = {
            k: v for k, v in astra_document.items() if k not in self._non_md_fields
        }
        return Document(
            page_content=astra_document[self.content_field],
            metadata=_metadata,
            id=astra_document["_id"],
        )

    @override
    def encode_filter(self, filter_dict: dict[str, Any]) -> dict[str, Any]:
        return filter_dict

    @override
    def metadata_key_to_field_identifier(self, md_key: str) -> str:
        return _flat_metadata_key_to_field_identifier(md_key)

    @override
    def encode_vectorize_sort(self, query_text: str) -> dict[str, Any]:
        raise ValueError(VECTORIZE_NOT_AVAILABLE_MSG)

    @override
    def encode_hybrid_sort(
        self,
        *,
        vector: list[float] | None,
        vectorize: str | None,
        lexical: str,
    ) -> dict[str, Any]:
        if vector is None or vectorize is not None:
            msg = "This codec's hybrid sort requires `vectorize=None` and a vector."
            raise ValueError(msg)
        return _astra_vector_encode_hybrid_sort(vector=vector, lexical=lexical)

    @property
    @override
    def default_collection_indexing_policy(self) -> dict[str, list[str]]:
        return {"deny": [self.content_field]}


class _FlatVectorizeVSDocumentCodec(_AstraDBVectorStoreDocumentCodec):
    """Codec for collections populated externally, with server-side embeddings.

    This codec manages document structured as a flat key-value map, with one
    field being the textual content and the other implicitly forming the "metadata".
    """

    server_side_embeddings = True
    content_field = VECTORIZE_FIELD_NAME

    def __init__(self, *, ignore_invalid_documents: bool, has_lexical: bool) -> None:
        """Initialize a new DefaultVectorizeVSDocumentCodec.

        Args:
            ignore_invalid_documents: if True, noncompliant inputs to `decode`
                are logged and a None is returned (instead of raising an exception).
            has_lexical: whether the codec should use the lexical field (hybrid search)
        """
        self.base_projection = {
            "_id": True,
            VECTOR_FIELD_NAME: False,
            VECTORIZE_FIELD_NAME: True,
        }
        self.full_projection = {"*": True}
        self.ignore_invalid_documents = ignore_invalid_documents
        self.has_lexical = has_lexical
        self._non_md_fields = {
            "_id",
            LEXICAL_FIELD_NAME,
            VECTOR_FIELD_NAME,
            VECTORIZE_FIELD_NAME,
            SIMILARITY_FIELD_NAME,
        }

    @override
    def encode(
        self,
        content: str,
        document_id: str,
        vector: list[float] | None,
        metadata: dict | None,
    ) -> dict[str, Any]:
        if vector is not None:
            msg = f"{VECTOR_REQUIRED_PREAMBLE_MSG}: {vector}"
            raise ValueError(msg)
        if VECTORIZE_FIELD_NAME in (metadata or {}):
            msg = FLATTEN_CONFLICT_MSG.format(field=VECTORIZE_FIELD_NAME)
            raise ValueError(msg)
        return {
            VECTORIZE_FIELD_NAME: content,
            "_id": document_id,
            **(metadata or {}),
            **({LEXICAL_FIELD_NAME: content} if self.has_lexical else {}),
        }

    @override
    def decode(self, astra_document: dict[str, Any]) -> Document | None:
        if VECTORIZE_FIELD_NAME not in astra_document and self.ignore_invalid_documents:
            invalid_doc_warning = (
                "Ignoring document with _id = "
                f"{astra_document.get('_id', '(no _id)')}. "
                "Reason: missing required fields."
            )
            warnings.warn(
                invalid_doc_warning,
                stacklevel=2,
            )
            return None
        _metadata = {
            k: v for k, v in astra_document.items() if k not in self._non_md_fields
        }
        return Document(
            page_content=astra_document[VECTORIZE_FIELD_NAME],
            metadata=_metadata,
            id=astra_document["_id"],
        )

    @override
    def encode_filter(self, filter_dict: dict[str, Any]) -> dict[str, Any]:
        return filter_dict

    @override
    def metadata_key_to_field_identifier(self, md_key: str) -> str:
        return _flat_metadata_key_to_field_identifier(md_key)

    @override
    def encode_vectorize_sort(self, query_text: str) -> dict[str, Any]:
        return _astra_generic_encode_vectorize_sort(query_text)

    @override
    def encode_hybrid_sort(
        self,
        *,
        vector: list[float] | None,
        vectorize: str | None,
        lexical: str,
    ) -> dict[str, Any]:
        if vectorize is None or vector is not None:
            msg = "This codec's hybrid sort requires `vector=None` and a vectorize."
            raise ValueError(msg)
        return _astra_vectorize_encode_hybrid_sort(vectorize=vectorize, lexical=lexical)

    @property
    @override
    def default_collection_indexing_policy(self) -> dict[str, list[str]]:
        # $vectorize cannot be de-indexed explicitly (the API manages it entirely).
        return {}



"""Loader for loading documents from DataStax Astra DB."""

from __future__ import annotations

import json
import logging
import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Callable,
    Iterator,
)

from langchain_core.document_loaders import BaseLoader
from langchain_core.documents import Document
from typing_extensions import override

from langchain_astradb.utils.astradb import (
    COMPONENT_NAME_LOADER,
    SetupMode,
    _AstraDBCollectionEnvironment,
)

if TYPE_CHECKING:
    from astrapy.authentication import TokenProvider

logger = logging.getLogger(__name__)

_NOT_SET = object()


class AstraDBLoader(BaseLoader):
    def __init__(
        self,
        collection_name: str,
        *,
        token: str | TokenProvider | None = None,
        api_endpoint: str | None = None,
        environment: str | None = None,
        namespace: str | None = None,
        filter_criteria: dict[str, Any] | None = None,
        projection: dict[str, Any] | None = _NOT_SET,  # type: ignore[assignment]
        limit: int | None = None,
        nb_prefetched: int = _NOT_SET,  # type: ignore[assignment]
        page_content_mapper: Callable[[dict], str] = json.dumps,
        metadata_mapper: Callable[[dict], dict[str, Any]] | None = None,
        ext_callers: list[tuple[str | None, str | None] | str | None] | None = None,
    ) -> None:
        """Load DataStax Astra DB documents.

        Args:
            collection_name: name of the Astra DB collection to use.
            token: API token for Astra DB usage, either in the form of a string
                or a subclass of `astrapy.authentication.TokenProvider`.
                If not provided, the environment variable
                ASTRA_DB_APPLICATION_TOKEN is inspected.
            api_endpoint: full URL to the API endpoint, such as
                `https://<DB-ID>-us-east1.apps.astra.datastax.com`. If not provided,
                the environment variable ASTRA_DB_API_ENDPOINT is inspected.
            environment: a string specifying the environment of the target Data API.
                If omitted, defaults to "prod" (Astra DB production).
                Other values are in `astrapy.constants.Environment` enum class.
            namespace: namespace (aka keyspace) where the collection resides.
                If not provided, the environment variable ASTRA_DB_KEYSPACE is
                inspected. Defaults to the database's "default namespace".
            filter_criteria: Criteria to filter documents.
            projection: Specifies the fields to return. If not provided, reads
                fall back to the Data API default projection.
            limit: a maximum number of documents to return in the read query.
            nb_prefetched: Max number of documents to pre-fetch.
                *IGNORED starting from v. 0.3.5: astrapy v1.0+ does not support it.*
            page_content_mapper: Function applied to collection documents to create
                the `page_content` of the LangChain Document. Defaults to `json.dumps`.
            metadata_mapper: Function applied to collection documents to create the
                `metadata` of the LangChain Document. Defaults to returning the
                 namespace, API endpoint and collection name.
            ext_callers: one or more caller identities to identify Data API calls
                in the User-Agent header. This is a list of (name, version) pairs,
                or just strings if no version info is provided, which, if supplied,
                becomes the leading part of the User-Agent string in all API requests
                related to this component.
        """
        astra_db_env = _AstraDBCollectionEnvironment(
            collection_name=collection_name,
            token=token,
            api_endpoint=api_endpoint,
            keyspace=namespace,
            environment=environment,
            setup_mode=SetupMode.OFF,
            ext_callers=ext_callers,
            component_name=COMPONENT_NAME_LOADER,
        )
        self.astra_db_env = astra_db_env
        self.filter = filter_criteria
        self._projection: dict[str, Any] | None = (
            projection if projection is not _NOT_SET else {"*": True}
        )
        # warning if 'prefetched' passed
        if nb_prefetched is not _NOT_SET:
            warnings.warn(
                (
                    "Parameter 'nb_prefetched' is not supported by the Data API "
                    "client and will be ignored in reading document."
                ),
                UserWarning,
                stacklevel=2,
            )

        self.limit = limit
        self.nb_prefetched = nb_prefetched
        self.page_content_mapper = page_content_mapper
        self.metadata_mapper = metadata_mapper or (
            lambda _: {
                "namespace": self.astra_db_env.database.keyspace,
                "api_endpoint": self.astra_db_env.database.api_endpoint,
                "collection": collection_name,
            }
        )

    def _to_langchain_doc(self, doc: dict[str, Any]) -> Document:
        return Document(
            page_content=self.page_content_mapper(doc),
            metadata=self.metadata_mapper(doc),
        )

    @override
    def lazy_load(self) -> Iterator[Document]:
        for doc in self.astra_db_env.collection.find(
            filter=self.filter,
            projection=self._projection,
            limit=self.limit,
            # prefetch: not available at the moment (silently ignored)
            # prefetched=self.nb_prefetched,
        ):
            yield self._to_langchain_doc(doc)

    async def aload(self) -> list[Document]:
        """Load data into Document objects."""
        return [doc async for doc in self.alazy_load()]

    @override
    async def alazy_load(self) -> AsyncIterator[Document]:
        async for doc in self.astra_db_env.async_collection.find(
            filter=self.filter,
            projection=self._projection,
            limit=self.limit,
            # prefetch: not available at the moment (silently ignored):
            # prefetched=self.nb_prefetched,
        ):
            yield self._to_langchain_doc(doc)



"""Utilities for AstraDB vector store autodetect mode."""

from __future__ import annotations

import logging
import warnings
from collections import Counter
from operator import itemgetter
from typing import (
    Any,
)

from langchain_astradb.utils.vector_store_codecs import (
    LEXICAL_FIELD_NAME,
    _AstraDBVectorStoreDocumentCodec,
    _DefaultVectorizeVSDocumentCodec,
    _DefaultVSDocumentCodec,
    _FlatVectorizeVSDocumentCodec,
    _FlatVSDocumentCodec,
)

logger = logging.getLogger(__name__)


def _detect_document_flatness(document: dict[str, Any]) -> bool | None:
    """Try to guess, when possible, if this document has metadata-as-a-dict or not."""
    _metadata = document.get("metadata")
    _vector = document.get("$vector")
    _regularfields = set(document.keys()) - {"_id", "$vector"}
    _regularfields_m = _regularfields - {"metadata"}
    # cannot determine if ...
    if _vector is None:
        return None
    # now a determination
    if isinstance(_metadata, dict) and _regularfields_m:
        return False
    if isinstance(_metadata, dict) and not _regularfields_m:
        # this document should not contribute to the survey
        return None
    str_regularfields = {
        k for k, v in document.items() if isinstance(v, str) if k in _regularfields
    }
    if str_regularfields:
        # Note: even if the only string field is "metadata"
        return True
    return None


def _detect_documents_flatness(documents: list[dict[str, Any]]) -> bool:
    flatness_survey = [_detect_document_flatness(document) for document in documents]
    n_flats = flatness_survey.count(True)
    n_deeps = flatness_survey.count(False)
    if n_flats > 0 and n_deeps > 0:
        msg = "Mixed document shapes detected on collection during autodetect."
        raise ValueError(msg)

    # in absence of clues, 0 < 0 is False and default is NON FLAT (i.e. native)
    return n_deeps < n_flats


def _detect_document_content_field(document: dict[str, Any]) -> str | None:
    """Try to guess the content field by inspecting the passed document.

    Prefer other fields than $lexical if available.
    """
    strlen_map = {
        k: len(v) for k, v in document.items() if k != "_id" if isinstance(v, str)
    }
    if not strlen_map:
        return None

    preferred_strlen_map = {k: v for k, v in strlen_map.items() if k != "$lexical"}
    if preferred_strlen_map:
        return max(preferred_strlen_map.items(), key=itemgetter(1))[0]
    return max(strlen_map.items(), key=itemgetter(1))[0]


def _detect_documents_content_field(
    documents: list[dict[str, Any]],
    requested_content_field: str,
) -> str:
    if requested_content_field == "*":
        # guess content_field by docs inspection
        content_fields = [
            _detect_document_content_field(document) for document in documents
        ]
        valid_content_fields = [cf for cf in content_fields if cf is not None]
        logger.info(
            "vector store autodetect: inferring content_field from %i documents",
            len(valid_content_fields),
        )
        cf_stats = Counter(valid_content_fields)
        if not cf_stats:
            msg = "Could not infer content_field name from sampled documents."
            raise ValueError(msg)
        return cf_stats.most_common(1)[0][0]

    return requested_content_field


def _detect_documents_have_lexical(documents: list[dict[str, Any]]) -> bool:
    return any(LEXICAL_FIELD_NAME in document for document in documents)


def _detect_document_codec(
    documents: list[dict[str, Any]],
    *,
    has_vectorize: bool,
    has_lexical: bool,
    ignore_invalid_documents: bool,
    norm_content_field: str,
) -> _AstraDBVectorStoreDocumentCodec:
    logger.info("vector store autodetect: inspecting %i documents", len(documents))
    # survey and determine flatness
    is_flat = _detect_documents_flatness(documents)
    logger.info("vector store autodetect: is_flat = %s", is_flat)

    final_content_field = _detect_documents_content_field(
        documents=documents,
        requested_content_field=norm_content_field,
    )
    logger.info(
        "vector store autodetect: final_content_field = %s", final_content_field
    )

    lexical_in_docs = _detect_documents_have_lexical(documents)
    logger.info(
        "vector store autodetect: lexical_in_docs = %s (has_lexical = %s)",
        lexical_in_docs,
        has_lexical,
    )
    if has_lexical and not lexical_in_docs:
        warnings.warn(
            "Collection supports lexical; however, autodetect encountered documents "
            "without a $lexical. The $lexical field will be set for new documents, "
            "but the pre-existing documents might not be reached by lexical search "
            "when running hybrid search mode.",
            stacklevel=2,
        )
    if not has_lexical and lexical_in_docs:
        warnings.warn(
            "Documents with $lexical encountered on a non-hybrid-capable "
            "collection. This inconsistency will be ignored moving forward "
            "(and new documents will not be written with any $lexical).",
            stacklevel=2,
        )

    if has_vectorize:
        if is_flat:
            return _FlatVectorizeVSDocumentCodec(
                ignore_invalid_documents=ignore_invalid_documents,
                has_lexical=has_lexical,
            )

        return _DefaultVectorizeVSDocumentCodec(
            ignore_invalid_documents=ignore_invalid_documents,
            has_lexical=has_lexical,
        )
    # no vectorize:
    if is_flat:
        return _FlatVSDocumentCodec(
            content_field=final_content_field,
            ignore_invalid_documents=ignore_invalid_documents,
            has_lexical=has_lexical,
        )
    return _DefaultVSDocumentCodec(
        content_field=final_content_field,
        ignore_invalid_documents=ignore_invalid_documents,
        has_lexical=has_lexical,
    )



"""Astra DB integration for LangChain."""

from astrapy.info import VectorServiceOptions

from langchain_astradb.cache import AstraDBCache, AstraDBSemanticCache
from langchain_astradb.chat_message_histories import AstraDBChatMessageHistory
from langchain_astradb.document_loaders import AstraDBLoader
from langchain_astradb.graph_vectorstores import AstraDBGraphVectorStore
from langchain_astradb.storage import AstraDBByteStore, AstraDBStore
from langchain_astradb.vectorstores import AstraDBVectorStore, AstraDBVectorStoreError

__all__ = [
    "AstraDBByteStore",
    "AstraDBCache",
    "AstraDBChatMessageHistory",
    "AstraDBGraphVectorStore",
    "AstraDBLoader",
    "AstraDBSemanticCache",
    "AstraDBStore",
    "AstraDBVectorStore",
    "AstraDBVectorStoreError",
    "VectorServiceOptions",
]
